from astropy.cosmology import Planck13 as cosmo
import numpy as np
import scipy
import george
import extinction


class LightCurve(object):
    """Light Curve class
    """
    def __init__(self, name, times, fluxes, flux_errs, filters,
                 zpt=0, mwebv=0, redshift=None, lim_mag=None,
                 obj_type=None):

        self.name = name
        self.times = times
        self.fluxes = fluxes
        self.flux_errs = flux_errs
        self.filters = filters
        self.zpt = zpt
        self.mwebv = mwebv
        self.redshift = redshift
        self.lim_mag = lim_mag
        self.obj_type = obj_type

        self.abs_mags = None
        self.abs_mags_err = None
        self.abs_lim_mag = None

    def sort_lc(self):
        gind = np.argsort(self.times)
        self.times = self.times[gind]
        self.fluxes = self.fluxes[gind]
        self.flux_errs = self.flux_errs[gind]
        self.filters = self.filters[gind]
        if self.abs_mags is not None:
            self.abs_mags = self.abs_mags[gind]
            self.abs_mags_err = self.abs_mags_err[gind]

    def find_peak(self, tpeak_guess):
        gind = np.where((np.abs(self.times-tpeak_guess) < 100.0) &
                        (self.fluxes/self.flux_errs > 3.0))
        if len(gind[0]) == 0:
            gind = np.where((np.abs(self.times - tpeak_guess) < 100.0))
        if self.abs_mags is not None:
            tpeak = self.times[gind][np.argmin(self.abs_mags[gind])]
        return tpeak

    def cut_lc(self, limit_before=70, limit_after=200):
        gind = np.where((self.times > -limit_before) &
                        (self.times < limit_after))
        self.times = self.times[gind]
        self.fluxes = self.fluxes[gind]
        self.flux_errs = self.flux_errs[gind]
        self.filters = self.filters[gind]
        if self.abs_mags is not None:
            self.abs_mags = self.abs_mags[gind]
            self.abs_mags_err = self.abs_mags_err[gind]

    def shift_lc(self, t0=0):
        self.times = self.times - t0

    def correct_time_dilation(self):
        self.times = self.times / (1.+self.redshift)

    def correct_extinction(self, wvs):
        alams = extinction.fm07(wvs, self.mwebv)
        for i, alam in enumerate(alams):
            gind = np.where(self.filters == str(i))
            self.abs_mags[gind] = self.abs_mags[gind] - alam

    def add_LC_info(self, zpt=27.5, mwebv=0.0, redshift=0.0,
                    lim_mag=25.0, obj_type='-'):
        self.zpt = zpt
        self.mwebv = mwebv
        self.redshift = redshift
        self.lim_mag = lim_mag
        self.obj_type = obj_type

    def get_abs_mags(self, replace_nondetections=True, mag_err_fill=1.0):
        """
        Convert flux into absolute magnitude

        Parameters
        ----------
        replace_nondetections : bool
            Replace nondetections with limiting mag.

        Returns
        -------
        self.abs_mags : list
            Absolute magnitudes

        Examples
        --------
        """
        k_correction = 2.5 * np.log10(1.+self.redshift)
        dist = cosmo.luminosity_distance([self.redshift]).value[0]  # returns dist in Mpc

        self.abs_mags = -2.5 * np.log10(self.fluxes) + self.zpt - 5. * \
            np.log10(dist*1e6/10.0) + k_correction
        # Sketchy way to calculate error - update later
        self.abs_mags_plus_err = -2.5 * np.log10(self.fluxes + self.flux_errs) + self.zpt - 5. * \
            np.log10(dist*1e6/10.0) + k_correction
        self.abs_mags_err = np.abs(self.abs_mags_plus_err - self.abs_mags)

        if replace_nondetections:
            abs_lim_mag = self.lim_mag - 5.0 * np.log10(dist * 1e6 / 10.0) + \
                            k_correction
            gind = np.where((np.isnan(self.abs_mags)) |
                            np.isinf(self.abs_mags) |
                            np.isnan(self.abs_mags_err) |
                            np.isinf(self.abs_mags_err) |
                            (self.abs_mags > self.lim_mag))

            self.abs_mags[gind] = abs_lim_mag
            self.abs_mags_err[gind] = mag_err_fill
        self.abs_lim_mag = abs_lim_mag

        return self.abs_mags, self.abs_mags_err

    def filter_names_to_numbers(self, filt_dict):
        for i, filt in enumerate(self.filters):
            self.filters[i] = filt_dict[filt]

    def make_dense_LC(self, nfilts):
        gp_mags = self.abs_mags - self.abs_lim_mag
        dense_fluxes = np.zeros((len(self.times), nfilts))
        dense_errs = np.zeros((len(self.times), nfilts))
        stacked_data = np.vstack([self.times, self.filters]).T
        x_pred = np.zeros((len(self.times)*nfilts, 2))
        kernel = np.var(gp_mags) * george.kernels.ExpSquaredKernel([100, 1], ndim=2)
        gp = george.GP(kernel)
        gp.compute(stacked_data, self.abs_mags_err)

        def neg_ln_like(p):
            gp.set_parameter_vector(p)
            return -gp.log_likelihood(gp_mags)

        def grad_neg_ln_like(p):
            gp.set_parameter_vector(p)
            return -gp.grad_log_likelihood(gp_mags)

        result = scipy.optimize.minimize(neg_ln_like,
                                         gp.get_parameter_vector(),
                                         jac=grad_neg_ln_like)
        gp.set_parameter_vector(result.x)
        for jj, time in enumerate(self.times):
            x_pred[jj*nfilts:jj*nfilts+nfilts, 0] = [time]*nfilts
            x_pred[jj*nfilts:jj*nfilts+nfilts, 1] = np.arange(nfilts)
        pred, pred_var = gp.predict(gp_mags, x_pred, return_var=True)

        for jj in np.arange(nfilts):
            gind = np.where(x_pred[:, 1] == jj)[0]
            dense_fluxes[:, int(jj)] = pred[gind] + self.abs_lim_mag
            dense_errs[:, int(jj)] = np.sqrt(pred_var[gind])
        self.dense_lc = np.dstack((dense_fluxes, dense_errs))
        gp.recompute()
        self.gp = gp
        self.gp_mags = gp_mags
        return gp, gp_mags
        # Need except statementgp.set_parameter_vector([1, 100, 1])



# lstm autoencoder recreate sequence
from argparse import ArgumentParser
from keras.models import Model
from keras.layers import Input, GRU, TimeDistributed
from keras.layers import Dense, concatenate
from keras.layers import RepeatVector
from keras.optimizers import Adam
import numpy as np
import matplotlib.pyplot as plt
import keras.backend as K
from keras.callbacks import EarlyStopping
import datetime
import os
import logging

now = datetime.datetime.now()
date = str(now.strftime("%Y-%m-%d"))

NEURON_N_DEFAULT = 100
ENCODING_N_DEFAULT = 10
N_EPOCH_DEFAULT = 1000


def customLoss(yTrue, yPred):
    """
    Custom loss which doesn't use the errors

    Parameters
    ----------
    yTrue : array
        True flux values
    yPred : array
        Predicted flux values
    """
    return K.mean(K.square(yTrue[:, :, 1:5] - yPred[:, :, :]))


def prep_input(input_lc_file, new_t_max=100.0, filler_err=1.0,
               save=False, load=False, outdir=None, prep_file=None):
    """
    Prep input file for fitting

    Parameters
    ----------
    input_lc_file : str
        True flux values
    new_t_max : float
        Predicted flux values
    filler_err : float
        Predicted flux values
    save : bool
        Predicted flux values
    load : bool
        Predicted flux values
    outdir : str
        Predicted flux values
    prep_file : str
        Predicted flux values

    Returns
    -------
    sequence : numpy.ndarray
        Array LC flux times, values and errors
    outseq : numpy.ndarray
        An array of LC flux values and limiting magnitudes
    ids : numpy.ndarray
        Array of SN names
    sequence_len : float
        Maximum length of LC values
    nfilts : int
        Number of filters in LC files
    """
    lightcurves = np.load(input_lc_file, allow_pickle=True)['lcs']
    lengths = []
    ids = []
    for lightcurve in lightcurves:
        lengths.append(len(lightcurve.times))
        ids.append(lightcurve.name)

    sequence_len = np.max(lengths)
    nfilts = np.shape(lightcurves[0].dense_lc)[1]
    nfiltsp1 = nfilts+1
    n_lcs = len(lightcurves)
    # convert from LC format to list of arrays
    sequence = np.zeros((n_lcs, sequence_len, nfilts*2+1))

    lms = []
    for i, lightcurve in enumerate(lightcurves):
        sequence[i, 0:lengths[i], 0] = lightcurve.times
        sequence[i, 0:lengths[i], 1:nfiltsp1] = lightcurve.dense_lc[:, :, 0]
        sequence[i, 0:lengths[i], nfiltsp1:] = lightcurve.dense_lc[:, :, 1]
        sequence[i, lengths[i]:, 0] = np.max(lightcurve.times)+new_t_max
        sequence[i, lengths[i]:, 1:nfiltsp1] = lightcurve.abs_lim_mag
        sequence[i, lengths[i]:, nfiltsp1:] = filler_err
        lms.append(lightcurve.abs_lim_mag)

    # Flip because who needs negative magnitudes
    sequence[:, :, 1:nfiltsp1] = -1.0 * sequence[:, :, 1:nfiltsp1]

    if load:
        prep_data = np.load(prep_file)
        bandmin = prep_data['bandmin']
        bandmax = prep_data['bandmax']
    else:
        bandmin = np.min(sequence[:, :, 1:nfiltsp1])
        bandmax = np.max(sequence[:, :, 1:nfiltsp1])

    sequence[:, :, 1:nfiltsp1] = (sequence[:, :, 1:nfiltsp1] - bandmin) \
        / (bandmax - bandmin)
    sequence[:, :, nfiltsp1:] = (sequence[:, :, nfiltsp1:]) \
        / (bandmax - bandmin)

    new_lms = np.reshape(np.repeat(lms, sequence_len), (len(lms), -1))

    outseq = np.reshape(sequence[:, :, 0], (len(sequence), sequence_len, 1)) * 1.0
    outseq = np.dstack((outseq, new_lms))
    if save:
        model_prep_file = outdir+'prep_'+date+'.npz'
        np.savez(model_prep_file, bandmin=bandmin, bandmax=bandmax)
        model_prep_file = outdir+'prep.npz'
        np.savez(model_prep_file, bandmin=bandmin, bandmax=bandmax)
    return sequence, outseq, ids, sequence_len, nfilts


def make_model(LSTMN, encodingN, maxlen, nfilts):
    """
    Make RAENN model

    Parameters
    ----------
    LSTMN : int
        Number of neurons to use in first/last layers
    encodingN : int
        Number of neurons to use in encoding layer
    maxlen : int
        Maximum LC length
    nfilts : int
        Number of filters in LCs

    Returns
    -------
    model : keras.models.Model
        RAENN model to be trained
    callbacks_list : list
        List of keras callbacks
    input_1 : keras.layer
        Input layer of RAENN
    encoded : keras.layer
        RAENN encoding layer
    """

    input_1 = Input((None, nfilts*2+1))
    input_2 = Input((maxlen, 2))

    encoder1 = GRU(LSTMN, return_sequences=True, activation='tanh', recurrent_activation='hard_sigmoid')(input_1)
    encoded = GRU(encodingN, return_sequences=False, activation='tanh',
                  recurrent_activation='hard_sigmoid')(encoder1)
    repeater = RepeatVector(maxlen)(encoded)
    merged = concatenate([repeater, input_2], axis=-1)
    decoder1 = GRU(LSTMN, return_sequences=True, activation='tanh', recurrent_activation='hard_sigmoid')(merged)
    decoder2 = TimeDistributed(Dense(nfilts, activation='tanh'),
                               input_shape=(None, 1))(decoder1)

    model = Model(input=[input_1, input_2], output=decoder2)

    new_optimizer = Adam(lr=1e-4, beta_1=0.9, beta_2=0.999,
                         decay=0)
    model.compile(optimizer=new_optimizer, loss=customLoss)

    es = EarlyStopping(monitor='val_loss', min_delta=0, patience=50,
                       verbose=0, mode='min', baseline=None,
                       restore_best_weights=True)

    callbacks_list = [es]
    return model, callbacks_list, input_1, encoded


def fit_model(model, callbacks_list, sequence, outseq, n_epoch):
    """
    Make RAENN model

    Parameters
    ----------
    model : keras.models.Model
        RAENN model to be trained
    callbacks_list : list
        List of keras callbacks
    sequence : numpy.ndarray
        Array LC flux times, values and errors
    outseq : numpy.ndarray
        An array of LC flux values and limiting magnitudes
    n_epoch : int
        Number of epochs to train for

    Returns
    -------
    model : keras.models.Model
        Trained keras model
    """
    model.fit([sequence, outseq], sequence, epochs=n_epoch,  verbose=1,
              shuffle=False, callbacks=callbacks_list, validation_split=0.33)
    return model


def test_model(sequence_test, model, lms, sequence_len, plot=True):
    outseq_test = np.reshape(sequence_test[:, :, 0], (len(sequence_test), sequence_len, 1))
    lms_test = np.reshape(np.repeat([lms], sequence_len), (len(sequence_test), -1))
    outseq_test = np.reshape(outseq_test[:, :, 0], (len(sequence_test), sequence_len, 1))
    outseq_test = np.dstack((outseq_test, lms_test))

    yhat = model.predict([sequence_test, outseq_test], verbose=1)
    if plot:
        plt.plot(sequence_test[0, :, 0], yhat[0, :, 1], color='grey')
        plt.plot(sequence_test[0, :, 0], sequence_test[0, :, 2], color='grey')
        plt.show()


def get_encoder(model, input_1, encoded):
    encoder = Model(input=input_1, output=encoded)
    return encoder


def get_decoder(model, encodingN):
    encoded_input = Input(shape=(None, (encodingN+2)))
    decoder_layer2 = model.layers[-2]
    decoder_layer3 = model.layers[-1]
    decoder = Model(input=encoded_input, output=decoder_layer3(decoder_layer2(encoded_input)))
    return decoder


def get_decodings(decoder, encoder, sequence, lms, encodingN, sequence_len, plot=True):

    if plot:
        for i in np.arange(len(sequence)):
            seq = np.reshape(sequence[i, :, :], (1, sequence_len, 9))
            encoding1 = encoder.predict(seq)[-1]
            encoding1 = np.vstack([encoding1]).reshape((1, 1, encodingN))
            repeater1 = np.repeat(encoding1, sequence_len, axis=1)
            out_seq = np.reshape(seq[:, :, 0], (len(seq), sequence_len, 1))
            lms_test = np.reshape(np.repeat(lms[i], sequence_len), (len(seq), -1))
            out_seq = np.dstack((out_seq, lms_test))

            decoding_input2 = np.concatenate((repeater1, out_seq), axis=-1)

            decoding2 = decoder.predict(decoding_input2)[0]

            plt.plot(seq[0, :, 0], seq[0, :, 1], 'green', alpha=1.0, linewidth=1)
            plt.plot(seq[0, :, 0], decoding2[:, 0], 'green', alpha=0.2, linewidth=10)
            plt.plot(seq[0, :, 0], seq[0, :, 2], 'red', alpha=1.0, linewidth=1)
            plt.plot(seq[0, :, 0], decoding2[:, 1], 'red', alpha=0.2, linewidth=10)
            plt.plot(seq[0, :, 0], seq[0, :, 3], 'orange', alpha=1.0, linewidth=1)
            plt.plot(seq[0, :, 0], decoding2[:, 2], 'orange', alpha=0.2, linewidth=10)
            plt.plot(seq[0, :, 0], seq[0, :, 4], 'purple', alpha=1.0, linewidth=1)
            plt.plot(seq[0, :, 0], decoding2[:, 3], 'purple', alpha=0.2, linewidth=10)
            plt.show()


def save_model(model, encodingN, LSTMN, model_dir='models/', outdir='./'):
    # make output dir
    model_dir = outdir + model_dir
    if not os.path.exists(model_dir):
        os.makedirs(model_dir)

    # serialize model to JSON
    model_json = model.to_json()
    with open(model_dir+"model_"+date+"_"+str(encodingN)+'_'+str(LSTMN)+".json", "w") as json_file:
        json_file.write(model_json)
    with open(model_dir+"model.json", "w") as json_file:
        json_file.write(model_json)
    # serialize weights to HDF5
    model.save_weights(model_dir+"model_"+date+"_"+str(encodingN)+'_'+str(LSTMN)+".h5")
    model.save_weights(model_dir+"model.h5")

    logging.info(f'Saved model to {model_dir}')


def save_encodings(model, encoder, sequence, ids, INPUT_FILE,
                   encodingN, LSTMN, N, sequence_len,
                   model_dir='encodings/', outdir='./'):

    # Make output directory
    model_dir = outdir + model_dir
    if not os.path.exists(model_dir):
        os.makedirs(model_dir)

    encodings = np.zeros((N, encodingN))
    for i in np.arange(N):
        seq = np.reshape(sequence[i, :, :], (1, sequence_len, 9))

        my_encoding = encoder.predict(seq)

        encodings[i, :] = my_encoding
        encoder.reset_states()

    encoder_sne_file = model_dir+'en_'+date+'_'+str(encodingN)+'_'+str(LSTMN)+'.npz'
    np.savez(encoder_sne_file, encodings=encodings, ids=ids, INPUT_FILE=INPUT_FILE)
    np.savez(model_dir+'en.npz', encodings=encodings, ids=ids, INPUT_FILE=INPUT_FILE)

    logging.info(f'Saved encodings to {model_dir}')


def main():
    parser = ArgumentParser()
    parser.add_argument('lcfile', type=str, help='Light curve file')
    parser.add_argument('--outdir', type=str, default='./products/',
                        help='Path in which to save the LC data (single file)')
    parser.add_argument('--plot', type=bool, default=False, help='Plot LCs')
    parser.add_argument('--neuronN', type=int, default=NEURON_N_DEFAULT, help='Number of neurons in hidden layers')
    parser.add_argument('--encodingN', type=int, default=ENCODING_N_DEFAULT,
                        help='Number of neurons in encoding layer')
    parser.add_argument('--n-epoch', type=int, dest='n_epoch',
                        default=N_EPOCH_DEFAULT,
                        help='Number of epochs to train for')

    args = parser.parse_args()

    sequence, outseq, ids, maxlen, nfilts = prep_input(args.lcfile, save=True, outdir=args.outdir)

    if args.plot:
        for s in sequence:
            plt.plot(s[:, 0], s[:, 1])
            plt.plot(s[:, 0], s[:, 2])
            plt.plot(s[:, 0], s[:, 3])
            plt.plot(s[:, 0], s[:, 4])
            plt.show()

    model, callbacks_list, input_1, encoded = make_model(args.neuronN,
                                                         args.encodingN,
                                                         maxlen, nfilts)
    model = fit_model(model, callbacks_list, sequence, outseq, args.n_epoch)
    encoder = get_encoder(model, input_1, encoded)

    # These comments used in testing, and sould be removed...
    # lms = outseq[:, 0, 1]
    # test_model(sequence_test, model, lm, maxlen, plot=True)
    # decoder = get_decoder(model, args.encodingN)
    # get_decodings(decoder, encoder, sequence, lms, args.encodingN, \
    #               maxlen, plot=False)

    if args.outdir[-1] != '/':
        args.outdir += '/'
    save_model(model, args.encodingN, args.neuronN, outdir=args.outdir)

    save_encodings(model, encoder, sequence, ids, args.lcfile,
                   args.encodingN, args.neuronN, len(ids), maxlen,
                   outdir=args.outdir)


if __name__ == '__main__':
    main()



import matplotlib.pyplot as plt
from feature_extraction import read_in_LC_files, feat_peaks, feat_rise_and_decline, feat_slope, feat_int


def get_peak(mydir, fname):
    not_done = True
    with open(mydir+fname) as f:
        while not_done:
            line = f.readline()
            if 'PEAKMJD' in line:
                pmjd = float(line.split(':')[-1])
                not_done = False
    return(pmjd)


test_dir = '/Volumes/My Passport/ps1_sne/all_lcs/output_lc/'
filename = 'PS1_PS1MD_PSc330023.snana.dat'
my_message = read_in_LC_files([test_dir+filename], ['PSc330023'])

filt_dict = {'g': 0, 'r': 1, 'i': 2, 'z': 3}

for my_lc in my_message:
    my_lc.add_LC_info(zpt=27.5, mwebv=0.0, redshift=0.1820,
                      lim_mag=25.0, obj_type='SNIa')
    my_lc.get_abs_mags()
    my_lc.sort_lc()

    pmjd = get_peak(test_dir, filename)
    pmjd = my_lc.find_peak(pmjd)
    my_lc.shift_lc(pmjd)
    my_lc.cut_lc(100)
    my_lc.filter_names_to_numbers(filt_dict)
    my_gp, gp_mags = my_lc.make_dense_LC(4)

plt.plot(my_lc.times, my_lc.abs_mags)
plt.plot(my_lc.times, my_lc.dense_lc[:, 0, 0])
plt.plot(my_lc.times, my_lc.dense_lc[:, 1, 0])
plt.plot(my_lc.times, my_lc.dense_lc[:, 2, 0])
plt.plot(my_lc.times, my_lc.dense_lc[:, 3, 0])
plt.errorbar(my_lc.times, my_lc.abs_mags, yerr=my_lc.abs_mags_err, line_style=None)
plt.show()


feat_peaks([my_lc])
t_rises_all, t_falls_all = feat_rise_and_decline([my_lc], [my_gp], [gp_mags], 1, 4)
print(t_rises_all, t_falls_all)
slopes = feat_slope([my_lc], [my_gp], [gp_mags])
print(slopes)
ints = feat_int([my_lc], [my_gp], [gp_mags])
print(ints)



import datetime
from argparse import ArgumentParser
import numpy as np
from sklearn import preprocessing
from sklearn.model_selection import LeaveOneOut
from sklearn.neighbors import KernelDensity
from sklearn.ensemble import RandomForestClassifier
import matplotlib.pyplot as plt
from sklearn.metrics import confusion_matrix
import pickle
import sys
import os
from astropy.table import QTable
from astropy.io import ascii

now = datetime.datetime.now()
date = str(now.strftime("%Y-%m-%d"))


def Gauss_resample(x, y, N):
    """
    Resample features based on Gaussian approximation

    Note we divide the covariance by 2!

    Parameters
    ----------
    X : numpy.ndarray
        Feature array
    y : numpy.ndarray
        Label array
    N : int
        Total samples to simulate (to be added to original sample)

    Returns
    -------
    newX : numpy.ndarray
        New Feature array
    newY : numpy.ndarray
        New label array
    """
    uys = np.unique(y)
    newX = np.zeros((int(N*len(uys)), np.size(x, axis=1)))
    newy = np.zeros((int(N*len(uys)), ))
    for i, uy in enumerate(uys):
        gind = np.where(y == uy)
        newX[i*N:i*N+len(gind[0]), :] = x[gind[0], :]
        newy[i*N:(i+1) * N] = uy
        cx = x[gind[0], :]
        mean = np.mean(cx, axis=0)
        cov = np.cov(cx, rowvar=False)
        newX[i * N + len(gind[0]):(i + 1) * N] = \
            np.random.multivariate_normal(mean, cov / 2., size=N - len(gind[0]))
    return newX, newy


def KDE_resample(x, y, N, bandwidth=0.5):
    """
    Resample features based on Kernel Density approximation

    Parameters
    ----------
    X : numpy.ndarray
        Feature array
    y : numpy.ndarray
        Label array
    N : int
        Total samples to simulate (to be added to original sample)

    Returns
    -------
    newX : numpy.ndarray
        New Feature array
    newY : numpy.ndarray
        New label array
    """
    uys = np.unique(y)
    newX = np.zeros((int(N*len(uys)), np.size(x, axis=1)))
    newy = np.zeros((int(N*len(uys)), ))
    for i, uy in enumerate(uys):
        gind = np.where(y == uy)
        newX[i * N:i * N + len(gind[0]), :] = x[gind[0], :]
        newy[i * N:(i + 1) * N] = uy
        cx = x[gind[0], :]
        kde = KernelDensity(kernel='gaussian', bandwidth=bandwidth).fit(cx)
        newX[i * N + len(gind[0]):(i + 1) * N] = kde.sample(n_samples=N - len(gind[0]))
    return newX, newy


def prep_data_for_classifying(featurefile, means, stds, whiten=True, verbose=False):
    """
    Resample features based on Kernel Density approximation

    Parameters
    ----------
    featurefile : str
        File with pre-processed features
    means : numpy.ndarray
        Means of features, used to whiten
    stds : numpy.ndarray
        St. dev of features, used to whiten
    whiten : bool
        Whiten features before classification
    verbose : bool
        Print if SNe fail

    Returns
    -------
    X : numpy.ndarray
        Feature array
    final_sn_names : numpy.ndarray
        Label array
    means : numpy.ndarray
        Means of features, used to whiten
    stds : numpy.ndarray
        St. dev of features, used to whiten
    feat_names : numpy.ndarray
        Array of feature names
    """
    feat_data = np.load(featurefile, allow_pickle=True)
    ids = feat_data['ids']
    features = feat_data['features']
    feat_names = feat_data['feat_names']

    X = []
    final_sn_names = []
    for sn_name in ids:
        gind = np.where(sn_name == ids)
        if verbose:
            if len(gind[0]) == 0:
                print('SN not found')
                sys.exit(0)
            if not np.isfinite(features[gind][0]).all():
                print('Warning: ', sn_name, ' has a non-finite feature')
        if X == []:
            X = features[gind][0]
        else:
            X = np.vstack((X, features[gind][0]))
        final_sn_names.append(sn_name)

    gind = np.where(np.isnan(X))
    if len(gind) > 0:
        X[gind[0], gind[1]] = means[gind[1]]
    if whiten:
        X = (X - means) / stds

    return X, final_sn_names, means, stds, feat_names


def prep_data_for_training(featurefile, metatable, whiten=True):
    """
    Resample features based on Kernel Density approximation

    Parameters
    ----------
    featurefile : str
        File with pre-processed features
    metatable : numpy.ndarray
        Table which must include: Object Name, Redshift, Type, Estimate
        Peak Time, and EBV_MW
    whiten : bool
        Whiten features before classification

    Returns
    -------
    X : numpy.ndarray
        Feature array
    y : numpy.ndarray
        Label array
    final_sn_names : numpy.ndarray
        Label array
    means : numpy.ndarray
        Means of features, used to whiten
    stds : numpy.ndarray
        St. dev of features, used to whiten
    feat_names : numpy.ndarray
        Array of feature names
    """
    feat_data = np.load(featurefile, allow_pickle=True)
    ids = feat_data['ids']
    features = feat_data['features']
    feat_names = feat_data['feat_names']
    metadata = np.loadtxt(metatable, dtype=str, usecols=(0, 2))
    sn_dict = {'SLSN': 0, 'SNII': 1, 'SNIIn': 2, 'SNIa': 3, 'SNIbc': 4}

    X = []
    y = []
    final_sn_names = []
    for sn_name, sn_type in metadata:
        gind = np.where(sn_name == ids)
        if 'SN' not in sn_type:
            continue
        else:
            sn_num = sn_dict[sn_type]

        if len(gind[0]) == 0:
            continue
        if not np.isfinite(features[gind][0]).all():
            continue

        if X == []:
            X = features[gind][0]
            y = sn_num
        else:
            X = np.vstack((X, features[gind][0]))
            y = np.append(y, sn_num)
        final_sn_names.append(sn_name)

    means = np.mean(X, axis=0)
    stds = np.std(X, axis=0)
    if whiten:
        X = preprocessing.scale(X)

    return X, y, final_sn_names, means, stds, feat_names


def main():
    parser = ArgumentParser()
    parser.add_argument('metatable', type=str, default='', help='Get training set labels')
    parser.add_argument('--featurefile', default='./products/feat.npz', type=str, help='Feature file')
    parser.add_argument('--outdir', type=str, default='./products/',
                        help='Path in which to save the LC data (single file)')
    parser.add_argument('--train', action='store_true',
                        help='Train classification model')
    parser.add_argument('--savemodel', action='store_true', help='Save output model, training on full set')
    parser.add_argument('--add-random', dest='add_random', type=bool, default=False,
                        help='Add random number as feature (for testing)')
    parser.add_argument('--calc-importance', dest='calc_importance', type=bool,
                        default=False, help='Calculate feature importance')
    parser.add_argument('--only-raenn', dest='only_raenn', type=bool, default=False, help='Use ony RAENN features')
    parser.add_argument('--not-raenn', dest='not_raenn', type=bool, default=False, help='Exclude RAENN features')
    parser.add_argument('--no-int', dest='no_int', type=bool, default=False,
                        help='Exclude integral features (for testing)')
    parser.add_argument('--resampling', dest='resampling', type=str, default='KDE',
                        help='Resampling methods. Either KDE or Gauss available')
    parser.add_argument('--modelfile', dest='modelfile', type=str,
                        default='model', help='Name of model file to save')
    parser.add_argument('--randomseed', type=int, default=42, help='Name of model file to save')
    parser.add_argument('--outfile', dest='outfile', type=str,
                        default='superprob', help='Name of probability table file')
    args = parser.parse_args()

    sn_dict = {'SLSN': 0, 'SNII': 1, 'SNIIn': 2, 'SNIa': 3, 'SNIbc': 4}

    if args.train:
        X, y, names, means, stds, feature_names = prep_data_for_training(args.featurefile, args.metatable)
        names = np.asarray(names, dtype=str)
        if args.only_raenn:
            gind = [i for i, feat in enumerate(feature_names) if 'raenn' in feat]
            X = X[:, gind]
            feature_names = feature_names[gind]

        if args.not_raenn:
            gind = [i for i, feat in enumerate(feature_names) if 'raenn' not in feat]
            X = X[:, gind]
            feature_names = feature_names[gind]

        if args.no_int:
            gind = [i for i, feat in enumerate(feature_names) if 'int' not in feat]
            X = X[:, gind]
            feature_names = feature_names[gind]

        if args.add_random:
            feature_names = np.append(feature_names, 'random')

        if not args.savemodel:
            loo = LeaveOneOut()
            y_pred = np.zeros(len(y))
            for train_index, test_index in loo.split(X):

                X_train, X_test = X[train_index], X[test_index]
                y_train, y_test = y[train_index], y[test_index]

                if args.resampling == 'Gauss':
                    X_res, y_res = Gauss_resample(X_train, y_train, 500)
                else:
                    X_res, y_res = KDE_resample(X_train, y_train, 500)

                new_ind = np.arange(len(y_res), dtype=int)
                np.random.shuffle(new_ind)
                X_res = X_res[new_ind]
                y_res = y_res[new_ind]

                if args.calc_importance:
                    X_res2, y_res2 = Gauss_resample(X_train, y_train, 500)
                    X_res2 = X_res2[:-40, :]
                    y_res2 = y_res2[:-40]

                if args.add_random:
                    X_res2, y_res2 = Gauss_resample(X_train, y_train, 500)
                    X_res2 = X_res2[:-40, :]
                    y_res2 = y_res2[:-40]
                    X_res = np.vstack((X_res.T, np.random.randn(len(X_res)))).T
                    X_res2 = np.vstack((X_res2.T, np.random.randn(len(X_res2)))).T
                    X_test = np.vstack((X_test.T, np.random.randn(len(X_test)))).T

                clf = RandomForestClassifier(n_estimators=400, max_depth=None,
                                             random_state=args.randomseed,
                                             criterion='gini',
                                             class_weight='balanced',
                                             max_features=None,
                                             oob_score=False)
                clf.fit(X_res, y_res)
                print(clf.predict_proba(X_test), y_test, names[test_index])

                if args.calc_importance:
                    feature_names = np.asarray(feature_names, dtype=str)
                    importances = clf.feature_importances_
                    indices = importances.argsort()[::-1]

                    print("Feature ranking:")

                    for f in range(X_res.shape[1]):
                        print(feature_names[indices[f]], importances[indices[f]])

                    plt.ylabel("Feature importances")
                    plt.bar(range(X_res.shape[1]), importances[indices],
                            color="grey", align="center")
                    plt.xticks(np.arange(len(importances))+0.5, feature_names[indices],
                               rotation=45, ha='right')
                    plt.show()
                y_pred[test_index] = np.argmax(clf.predict_proba(X_test))
            cnf_matrix = confusion_matrix(y, y_pred)
            print(cnf_matrix)
        if args.savemodel:
            if args.resampling == 'Gauss':
                X_res, y_res = Gauss_resample(X, y, 500)
            else:
                X_res, y_res = KDE_resample(X, y, 500)

            new_ind = np.arange(len(y_res), dtype=int)
            np.random.shuffle(new_ind)
            X_res = X_res[new_ind]
            y_res = y_res[new_ind]

            clf = RandomForestClassifier(n_estimators=350, max_depth=None,
                                         random_state=args.randomseed, criterion='gini', class_weight='balanced',
                                         max_features=None, oob_score=False)
            clf.fit(X_res, y_res)

            # save the model to disk
            if not os.path.exists(args.outdir):
                os.makedirs(args.outdir)
            if args.outdir[-1] != '/':
                args.outdir += '/'
            pickle.dump([clf, means, stds], open(args.outdir+args.modelfile+'_'+date+'.sav', 'wb'))
            pickle.dump([clf, means, stds], open(args.outdir+args.modelfile+'.sav', 'wb'))

    else:
        info = pickle.load(open(args.modelfile, 'rb'))
        loaded_model = info[0]
        means = info[1]
        stds = info[2]
        X, names, means, stds, feature_names = prep_data_for_classifying(args.featurefile, means, stds)
        names = np.asarray(names, dtype=str)
        probabilities = np.zeros((len(names), len(sn_dict)))
        for i, name in enumerate(names):
            probabilities[i] = loaded_model.predict_proba([X[i]])[0]
        probability_table = QTable(np.vstack((names, probabilities.T)).T,
                                   names=['Event Name', *sn_dict],
                                   meta={'name': 'SuperRAENN probabilities'})

        # save the model to disk
        if not os.path.exists(args.outdir):
            os.makedirs(args.outdir)
        if args.outdir[-1] != '/':
            args.outdir += '/'
        ascii.write(probability_table, args.outdir+args.outfile+'.tex',
                    format='latex', overwrite=True)


if __name__ == '__main__':
    main()



import numpy as np
from .lc import LightCurve
from .raenn import prep_input, get_decoder, get_decodings
import argparse
from keras.models import model_from_json, Model
from keras.layers import Input
import datetime
import os

now = datetime.datetime.now()
date = str(now.strftime("%Y-%m-%d"))


def str2bool(v):
    """
    Helper function to turn strings to bool

    """
    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 't', 'y', '1', 'True'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0', 'False'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')


def read_in_LC_files(input_files, obj_names, style='SNANA'):
    """
    Read in LC files and convert to LC object

    Parameters
    ----------
    input_files : list
        List of LC file names, to be read in.
    obj_names : list
        List of SNe names, should be same length as input_files
    style : string
        Style of LC files. Assumes SNANA

    Returns
    -------
    lcs : list
        list of Light Curve objects

    Examples
    --------
    """
    LC_list = []
    if style == 'SNANA':
        for i, input_file in enumerate(input_files):
            t, f, filts, err = np.genfromtxt(input_file,
                                             usecols=(1, 4, 2, 5), skip_header=18,
                                             skip_footer=1, unpack=True, dtype=str)
            t = np.asarray(t, dtype=float)
            f = np.asarray(f, dtype=float)
            err = np.asarray(err, dtype=float)

            sn_name = obj_names[i]
            new_LC = LightCurve(sn_name, t, f, err, filts)
            LC_list.append(new_LC)
    else:
        raise ValueError('Sorry, you need to specify a data style.')
    return LC_list


def feat_from_raenn(data_file, model_base=None,
                    prep_file=None, plot=False):
    """
    Calculate RAENN features

    Parameters
    ----------
    data_file : str
        Name of data file with light curves
    model_base : str
        Name of RAENN model file
    prep_file : str
        Name of file which encodes the feature prep

    Returns
    -------
    encodings : numpy.ndarray
        Array of object IDs (strings)

    TODO
    ----
    - prep file seems unnecessary
    """
    sequence, outseq, ids, maxlen, nfilts = prep_input(data_file, load=True, prep_file=prep_file)
    model_file = model_base + '.json'
    model_weight_file = model_base+'.h5'
    with open(model_file, 'r') as f:
        model = model_from_json(f.read())
    model.load_weights(model_weight_file)

    encodingN = model.layers[2].output_shape[1]
    original_input = Input(shape=(None, nfilts*2+1))
    encoded = model.layers[2]
    encoded1 = model.layers[1]
    encoder = Model(input=original_input, output=encoded(encoded1(original_input)))

    if plot:
        decoder = get_decoder(model, encodingN)
        lms = outseq[:, 0, 1]
        sequence_len = maxlen
        print(lms)
        get_decodings(decoder, encoder, sequence, lms, encodingN, sequence_len)

    encodings = np.zeros((len(ids), encodingN))
    for i in np.arange(len(ids)):
        inseq = np.reshape(sequence[i, :, :], (1, maxlen, nfilts*2+1))
        my_encoding = encoder.predict(inseq)
        encodings[i, :] = my_encoding
        encoder.reset_states()
    return encodings


def feat_peaks(input_lcs):
    """
    Extract peak magnitudes from GP LCs

    Parameters
    ----------
    input_lcs : list
        List of LC objects

    Returns
    -------
    peaks : list
        Peaks from each LC filter

    Examples
    --------
    """
    peaks = []
    for input_lc in input_lcs:
        peaks.append(np.nanmin(input_lc.dense_lc[:, :, 0], axis=0))
    return peaks


def feat_rise_and_decline(input_lcs, n_mag, nfilts=4):

    t_falls_all = []
    t_rises_all = []

    for i, input_lc in enumerate(input_lcs):
        gp = input_lc.gp
        gp_mags = input_lc.gp_mags
        t_falls = []
        t_rises = []
        for j in np.arange(nfilts):
            new_times = np.linspace(-100, 100, 500)
            x_stacked = np.asarray([new_times, [j] * 500]).T
            pred, var = gp.predict(gp_mags, x_stacked)

            max_ind = np.nanargmin(pred)
            max_mag = pred[max_ind]
            max_t = new_times[max_ind]
            trise = np.where((new_times < max_t) & (pred > (max_mag + n_mag)))
            tfall = np.where((new_times > max_t) & (pred > (max_mag + n_mag)))
            if len(trise[0]) == 0:
                trise = np.max(new_times) - max_t
            else:
                trise = max_t - new_times[trise][-1]
            if len(tfall[0]) == 0:
                tfall = max_t - np.min(new_times)
            else:
                tfall = new_times[tfall][0] - max_t

            t_falls.append(tfall)
            t_rises.append(trise)
        t_falls_all.append(t_falls)
        t_rises_all.append(t_rises)
    return t_rises_all, t_falls_all


def feat_slope(input_lcs, t_min_lim=10, t_max_lim=30, nfilts=4):
    slopes_all = []
    for i, input_lc in enumerate(input_lcs):
        gp = input_lc.gp
        gp_mags = input_lc.gp_mags
        slopes = []
        for j in np.arange(nfilts):
            new_times = np.linspace(-100, 100, 500)
            x_stacked = np.asarray([new_times, [j] * 500]).T
            pred, var = gp.predict(gp_mags, x_stacked)
            max_ind = np.nanargmin(pred)
            max_t = new_times[max_ind]
            new_times = new_times - max_t
            lc_grad = np.gradient(pred, new_times)
            gindmean = np.where((new_times > t_min_lim) & (new_times < t_max_lim))
            slopes.append(np.nanmedian(lc_grad[gindmean]))
        slopes_all.append(slopes)
    return slopes_all


def feat_int(input_lcs, nfilts=4):
    ints_all = []
    for i, input_lc in enumerate(input_lcs):
        gp = input_lc.gp
        gp_mags = input_lc.gp_mags
        ints = []
        for j in np.arange(nfilts):
            new_times = np.linspace(-100, 100, 500)
            x_stacked = np.asarray([new_times, [j] * 500]).T
            pred, var = gp.predict(gp_mags, x_stacked)
            ints.append(np.trapz(pred))

        ints_all.append(ints)
    return ints_all


def save_features(features, ids, feat_names, outputfile, outdir):
    # make output dir
    outputfile = outputfile+'.npz'
    outputfile = outdir + outputfile
    if not os.path.exists(outdir):
        os.makedirs(outdir)

    np.savez(outputfile, features=features, ids=ids, feat_names=feat_names)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('lcfile', type=str, help='Light curve file')
    parser.add_argument('--outdir', type=str, default='./products/',
                        help='Path in which to save the LC data (single file)')
    parser.add_argument('--plot', type=str2bool, default=False, help='Plot LCs, for testing')
    parser.add_argument('--model-base', type=str, dest='model_base', default='./products/models/model', help='...')
    parser.add_argument('--get-feat-raenn', type=str2bool, dest='get_feat_raenn', default=True, help='...')
    parser.add_argument('--get-feat-peaks', type=str2bool, dest='get_feat_peaks', default=True, help='...')
    parser.add_argument('--get-feat-rise-decline-1', type=str2bool,
                        dest='get_feat_rise_decline1', default=True,
                        help='...')
    parser.add_argument('--get-feat-rise-decline-2', type=str2bool,
                        dest='get_feat_rise_decline2', default=True,
                        help='...')
    parser.add_argument('--get-feat-rise-decline-3', type=str2bool,
                        dest='get_feat_rise_decline3', default=True,
                        help='...')
    parser.add_argument('--get-feat-slope', type=str2bool, dest='get_feat_slope', default=True, help='...')
    parser.add_argument('--get-feat-int', type=str2bool, dest='get_feat_int', default=True, help='...')
    parser.add_argument('--prep-file', type=str, dest='prep_file', default='./products/prep.npz', help='...')
    parser.add_argument('--outfile', type=str, dest='outfile', default='feat', help='...')

    args = parser.parse_args()
    features = []

    input_lcs = np.load(args.lcfile, allow_pickle=True)['lcs']
    ids = []
    feat_names = []
    for input_lc in input_lcs:
        ids.append(input_lc.name)
    if args.get_feat_raenn:
        feat = feat_from_raenn(args.lcfile, model_base=args.model_base,
                               prep_file=args.prep_file, plot=args.plot)
        if features != []:
            features = np.hstack((features, feat))
        else:
            features = feat
        for i in np.arange(np.shape(feat)[-1]):
            feat_names.append('raenn'+str(i))
        print('RAENN feat done')

    if args.get_feat_peaks:
        feat = feat_peaks(input_lcs)
        if features != []:
            features = np.hstack((features, feat))
        else:
            features = feat
        for i in np.arange(np.shape(feat)[-1]):
            feat_names.append('peak'+str(i))
        print('peak feat done')

    if args.get_feat_rise_decline1:
        feat1, feat2 = feat_rise_and_decline(input_lcs, 1)
        if features != []:
            features = np.hstack((features, feat1))
            features = np.hstack((features, feat2))
        else:
            features = np.hstack((feat1, feat2))
        for i in np.arange(np.shape(feat)[-1]):
            feat_names.append('rise1'+str(i))
        for i in np.arange(np.shape(feat)[-1]):
            feat_names.append('decline1'+str(i))
        print('dur1 feat done')

    if args.get_feat_rise_decline2:
        feat1, feat2 = feat_rise_and_decline(input_lcs, 2)
        if features != []:
            features = np.hstack((features, feat1))
            features = np.hstack((features, feat2))
        else:
            features = np.hstack((feat1, feat2))
        for i in np.arange(np.shape(feat)[-1]):
            feat_names.append('rise2'+str(i))
        for i in np.arange(np.shape(feat)[-1]):
            feat_names.append('decline2'+str(i))
        print('dur2 feat done')

    if args.get_feat_rise_decline3:
        feat1, feat2 = feat_rise_and_decline(input_lcs, 3)
        if features != []:
            features = np.hstack((features, feat1))
            features = np.hstack((features, feat2))
        else:
            features = np.hstack((feat1, feat2))
        for i in np.arange(np.shape(feat)[-1]):
            feat_names.append('rise3'+str(i))
        for i in np.arange(np.shape(feat)[-1]):
            feat_names.append('decline3'+str(i))
        print('dur3 feat done')

    if args.get_feat_slope:
        feat = feat_slope(input_lcs)
        if features != []:
            features = np.hstack((features, feat))
        else:
            features = feat
        for i in np.arange(np.shape(feat)[-1]):
            feat_names.append('slope'+str(i))
        print('slope feat done')

    if args.get_feat_int:
        feat = feat_int(input_lcs)
        if features != []:
            features = np.hstack((features, feat))
        else:
            features = feat
        for i in np.arange(np.shape(feat)[-1]):
            feat_names.append('int'+str(i))
        print('int feat done')

    if args.outdir[-1] != '/':
        args.outdir += '/'
    save_features(features, ids, feat_names, args.outfile+'_'+date, outdir=args.outdir)
    save_features(features, ids, feat_names, args.outfile, outdir=args.outdir)


if __name__ == '__main__':
    main()



from argparse import ArgumentParser
import numpy as np
from .feature_extraction import read_in_LC_files
import logging
import datetime
import os

logging.basicConfig(format='%(asctime)s %(message)s', datefmt='%Y-%m-%d %H:%M:%S', level=logging.INFO)

# Default fitting parameters
DEFAULT_ZPT = 27.5
DEFAULT_LIM_MAG = 25.0


def read_in_meta_table(metatable):
    """
    Read in the metatable file

    Parameters
    ----------
    metatable : str
        Name of metatable file
    Returns
    -------
    obj : numpy.ndarray
        Array of object IDs (strings)
    redshift : numpy.ndarray
        Array of redshifts
    obj_type : numpy.ndarray
        Array of SN spectroscopic types
    my_peak : numpy.ndarray
        Array of best-guess peak times
    ebv : numpy.ndarray
        Array of MW ebv values

    Todo
    ----------
    Make metatable more flexible
    """
    obj, redshift, obj_type, \
        my_peak, ebv = np.loadtxt(metatable, unpack=True, dtype=str, delimiter=' ')
    redshift = np.asarray(redshift, dtype=float)
    my_peak = np.asarray(my_peak, dtype=float)
    ebv = np.asarray(ebv, dtype=float)

    return obj, redshift, obj_type, my_peak, ebv


def save_lcs(lc_list, output_dir):
    """
    Save light curves as a lightcurve object

    Parameters
    ----------
    lc_list : list
        list of light curve files
    output_dir : Output directory of light curve file

    Todo:
    ----------
    - Add option for LC file name
    """
    now = datetime.datetime.now()
    date = str(now.strftime("%Y-%m-%d"))
    file_name = 'lcs_' + date + '.npz'

    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    if output_dir[-1] != '/':
        output_dir += '/'

    output_file = output_dir + file_name
    np.savez(output_file, lcs=lc_list)
    # Also easy save to latest
    np.savez(output_dir+'lcs.npz', lcs=lc_list)

    logging.info(f'Saved to {output_file}')


def main():
    """
    Preprocess the LC files
    """

    # Create argument parser
    parser = ArgumentParser()
    parser.add_argument('datadir', type=str, help='Directory of LC files')
    parser.add_argument('metatable', type=str,
                        help='Metatable containing each object, redshift, peak time guess, mwebv, object type')
    parser.add_argument('--zpt', type=float, default=DEFAULT_ZPT, help='Zero point of LCs')
    parser.add_argument('--lm', type=float, default=DEFAULT_LIM_MAG, help='Survey limiting magnitude')
    parser.add_argument('--outdir', type=str, default='./products/',
                        help='Path in which to save the LC data (single file)')
    args = parser.parse_args()

    objs, redshifts, obj_types, peaks, ebvs = read_in_meta_table(args.metatable)

    # Grab all the LC files in the input directory
    file_names = []
    for obj in objs:
        file_name = args.datadir + 'PS1_PS1MD_' + obj + '.snana.dat'
        file_names.append(file_name)

    # Create a list of LC objects from the data files
    lc_list = read_in_LC_files(file_names, objs)

    # This needs to be redone when retrained
    # TODO: Need to change this whenever you retrain...
    filt_dict = {'g': 0, 'r': 1, 'i': 2, 'z': 3}
    wvs = np.asarray([5460, 6800, 7450, 8700])

    # Update the LC objects with info from the metatable
    my_lcs = []
    for i, my_lc in enumerate(lc_list):
        my_lc.add_LC_info(zpt=args.zpt, mwebv=ebvs[i],
                          redshift=redshifts[i], lim_mag=args.lm,
                          obj_type=obj_types[i])
        my_lc.get_abs_mags()
        my_lc.sort_lc()
        pmjd = my_lc.find_peak(peaks[i])
        my_lc.shift_lc(pmjd)
        my_lc.correct_time_dilation()
        my_lc.filter_names_to_numbers(filt_dict)
        my_lc.correct_extinction(wvs)
        my_lc.cut_lc()
        my_lc.make_dense_LC(4)
        my_lcs.append(my_lc)
    save_lcs(my_lcs, args.outdir)


if __name__ == '__main__':
    main()
